import numpy as np
from scipy.sparse import csr_matrix, hstack, vstack
from sklearn.base import clone
from sklearn.utils import validation
from sklearn.exceptions import NotFittedError
from ..problem_transform import ClassifierChain


class ProbabilisticClassifierChain(ClassifierChain):
    """
    Probabilistic Classifier Chain for Multi-Label Classification

    This class implements a Probabilistic Classifier Chain (PCC), an extension of
    Jesse Read's Classifier Chains. It learns a chain of classifiers, each predicting
    a label conditioned on the input features and the predictions of preceding classifiers
    in the chain. This approach models joint label distributions to capture label correlations.

    Each classifier in the chain is trained on an augmented input space that includes the
    original features (X) and the predictions of all previous classifiers in the chain.

    The implementation adapted and changed from:
    https://github.com/ChristianSch/skml/blob/master/skml/problem_transformation/probabilistic_classifier_chain.py

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        A scikit-learn compatible base classifier. This classifier is used as the base model
        for each step in the classifier chain.
    require_dense : [bool, bool], optional
        Indicates whether the base classifier requires dense representations for input features
        and label matrices in fit/predict. If not provided, it defaults to using sparse
        representations unless the base classifier is an instance of
        :class:`~skmultilearn.base.MLClassifierBase`, in which case dense representations are used.
    order : List[int], permutation of ``range(n_labels)``, optional
        the order in which the chain should go through labels, the default is ``range(n_labels)``

    Attributes
    ----------
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `n_labels`
        A list of classifiers, one for each label, trained per partition as per the chain order.

    References
    ----------
    If using this implementation, please cite the scikit-multilearn library and the relevant paper:

    .. code-block:: bibtex

        @inproceedings{Dembczynski2010BayesOM,
          title={Bayes Optimal Multilabel Classification via Probabilistic Classifier Chains},
          author={Krzysztof Dembczynski and Weiwei Cheng and Eyke H{\"u}llermeier},
          booktitle={International Conference on Machine Learning},
          year={2010},
          url={https://api.semanticscholar.org/CorpusID:6418797}
        }

    Examples
    --------
    An example of using Probabilistic Classifier Chain with an :class:`sklearn.svm.SVC`:

    .. code-block:: python

        from skmultilearn.problem_transform import ProbabilisticClassifierChain
        from sklearn.svm import SVC

        # Initialize Probabilistic Classifier Chain with an SVM classifier
        classifier = ProbabilisticClassifierChain(
            classifier = SVC(probability=True),
            require_dense = [False, True]
        )

        # Train
        classifier.fit(X_train, y_train)

        # Predict
        predictions = classifier.predict(X_test)

    To optimize the classifier chain with grid search, one can vary both the base classifier
    and its parameters:

    .. code-block:: python

        from skmultilearn.problem_transform import ProbabilisticClassifierChain
        from sklearn.model_selection import GridSearchCV
        from sklearn.naive_bayes import MultinomialNB
        from sklearn.svm import SVC

        parameters = [
            {
                'classifier': [MultinomialNB()],
                'classifier__alpha': [0.7, 1.0],
            },
            {
                'classifier': [SVC(probability=True)],
                'classifier__kernel': ['rbf', 'linear'],
            },
        ]

        clf = GridSearchCV(ProbabilisticClassifierChain(), parameters, scoring='accuracy')
        clf.fit(X, y)

        print(clf.best_params_, clf.best_score_)

    """

    def __init__(self, classifier=None, require_dense=None, order=None):
        super(ProbabilisticClassifierChain, self).__init__(classifier, require_dense)
        self.order = order
        self.copyable_attrs = ["classifier", "require_dense", "order"]

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        validation.check_is_fitted(self, 'classifiers_')
        X_extended = self._ensure_input_format(
            X, sparse_format="csc", enforce_sparse=True
        )

        y_pred = []
        N_instances = X_extended.shape[0]

        for n in range(N_instances):
            x = X_extended[n].reshape(1, -1)
            y_out = None
            p_max = 0

            for b in range(2 ** self._label_count):
                p = np.zeros((1, self._label_count))
                y = np.array(list(map(int, np.binary_repr(b, width=self._label_count))))

                for i, c in enumerate(self.classifiers_):
                    if i == 0:
                        p[0, i] = c.predict_proba(self._ensure_input_format(x))[0][y[i]]
                    else:
                        y_slice = csr_matrix(y[:i].reshape(1, -1))
                        stacked = hstack((x, y_slice)).reshape(1, -1)
                        p[0, i] = c.predict_proba(self._ensure_input_format(stacked))[0][y[i]]

                pp = np.prod(p)

                if pp > p_max:
                    y_out = y
                    p_max = pp

            y_pred.append(y_out)

        return csr_matrix(y_pred)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """
        validation.check_is_fitted(self, 'classifiers_')
        X_extended = self._ensure_input_format(
            X, sparse_format="csc", enforce_sparse=True
        )

        results = []
        N_instances = X_extended.shape[0]

        for n in range(N_instances):
            x = X_extended[n].reshape(1, -1)
            y_out = None
            p_max = 0

            for b in range(2 ** self._label_count):
                p = np.zeros((1, self._label_count))
                y = np.array(list(map(int, np.binary_repr(b, width=self._label_count))))

                for i, c in enumerate(self.classifiers_):
                    if i == 0:
                        p[0, i] = c.predict_proba(self._ensure_input_format(x))[0][y[i]]
                    else:
                        y_slice = csr_matrix(y[:i].reshape(1, -1))
                        stacked = hstack([x, y_slice]).reshape(1, -1)
                        p[0, i] = c.predict_proba(self._ensure_input_format(stacked))[0][y[i]]

                pp = np.prod(p)

                if pp > p_max:
                    y_out = csr_matrix(p)
                    p_max = pp

            results.append(y_out)

        y_proba = vstack(results)
        return y_proba



import copy
import numpy as np

from scipy.sparse import hstack, issparse, lil_matrix
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier

from ..base.base import MLClassifierBase
from ..base.problem_transformation import ProblemTransformationBase


class InstanceBasedLogisticRegression(ProblemTransformationBase):
    """Combining Instance-Based Learning and Logistic Regression

    This idea is put into practice by means of a learning algorithm
    that realizes instance-based classification as
    logistic regression, using the information coming from the neighbors
    of an instance x as a “feature”.

    The first classifier layer is filled with K-Nearest Neighbor models,
    while the second classifier layer is filled with another classifiers,
    Logistic Regression as default.

    Parameters
    ----------
    require_dense : [bool, bool], optional
        whether the base classifier requires dense representations
        for input features and classes/labels matrices in fit/predict.
        If value not provided, sparse representations are used if base classifier is
        an instance of :class:`~skmultilearn.base.MLClassifierBase` and dense otherwise.

    Attributes
    ----------
    model_count_ : int
        number of trained models, in this classifier equal to `n_labels`
    partition_ : List[List[int]], shape=(`model_count_`,)
        list of lists of label indexes, used to index the output space matrix, set in :meth:`_generate_partition`
        via :meth:`fit`
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
        list of classifiers trained per partition, set in :meth:`fit`
    knn_layer_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
        list of classifiers trained per partition in first layer, set in :meth:`fit`

    References
    ----------
    If used, please cite the scikit-multilearn library and the relevant paper:

    .. code-block:: bibtex

        @article{cheng2009combining,
            title={Combining instance-based learning and logistic regression for multilabel classification},
            author={Cheng, Weiwei and H{\"u}llermeier, Eyke},
            journal={Machine Learning},
            volume={76},
            number={2-3},
            pages={211--225},
            year={2009},
            publisher={Springer}
        }

    Examples
    --------
    An example use case for Classifier Chains
    with an :class:`sklearn.svm.SVC` base classifier which supports sparse input:

    .. code-block:: python

        from skmultilearn.problem_transform import InstanceBasedLogisticRegression
        from sklearn.svm import SVC

        # initialize Instance-Based Learning and Logistic Regression multi-label classifier
        # with an SVM classifier
        # SVM in scikit only supports the X matrix in sparse representation

        classifier = InstanceBasedLogisticRegression(
            classifier = SVC(),
            require_dense = [False, True]
        )

        # train
        classifier.fit(X_train, y_train)

        # predict
        predictions = classifier.predict(X_test)

    Another way to use this classifier is to select the best scenario from a set of single-label classifiers used
    with Instance-Based Learning and Logistic Regression, this can be done using cross validation grid search. In
    the example below, the model with highest accuracy results is selected from either a
    :class:`sklearn.naive_bayes.MultinomialNB` or :class:`sklearn.svm.SVC` base classifier, alongside with best
    parameters for that base classifier.

    .. code-block:: python

        from skmultilearn.problem_transform import InstanceBasedLogisticRegression
        from sklearn.model_selection import GridSearchCV
        from sklearn.naive_bayes import MultinomialNB
        from sklearn.svm import SVC

        parameters = [
            {
                'classifier': [MultinomialNB()],
                'classifier__alpha': [0.7, 1.0],
            },
            {
                'classifier': [SVC()],
                'classifier__kernel': ['rbf', 'linear'],
            },
        ]

        # Create a ClassificationHeterogeneousFeature instance with GridSearchCV
        clf = GridSearchCV(InstanceBasedLogisticRegression(), parameters, scoring='accuracy')
        clf.fit(X, y)

        print(clf.best_params_, clf.best_score_)

        # Example output
        # {'classifier': MultinomialNB(alpha=0.7), 'classifier__alpha': 0.7} 0.18
    """

    def __init__(self, classifier=LogisticRegression(), require_dense=None):
        super(InstanceBasedLogisticRegression, self).__init__(classifier, require_dense)
        self.knn_classifier = KNeighborsClassifier(n_neighbors=10, n_jobs=-1)
        self.knn_layer_ = []

    def _generate_partition(self, X, y):
        """Partitions the label space into singletons
        Sets `self.partition_` (list of single item lists) and `self.model_count_` (equal to number of labels).
        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            not used, only for API compatibility
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `int`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        self.partition_ = list(range(y.shape[1]))
        self.model_count_ = y.shape[1]

    def _concatenate_class_membership(self, X, class_membership):
        """Concatenate original features and instance-based information from first layer

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        class_membership : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_labels)
            label predict probabilities from X
        """
        return hstack([X, class_membership]).tolil()

    def _get_class_membership(self, classifiers, X):
        """Extract instance-based information from original data X

        Parameters
        ----------
        classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
            list of pretrained classifiers per partition based on X
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        """
        result = lil_matrix((X.shape[0], self._label_count), dtype="float")

        for label_assignment, classifier in zip(self.partition_, classifiers):
            # n_samples x n_classes, where n_classes = [0, 1] - 1 is the probability of
            # the label being assigned
            result[:, label_assignment] = self._ensure_multi_label_from_single_class(
                classifier.predict_proba(self._ensure_input_format(X))
            )[
                :, 1
            ]  # probability that label is assigned

        return result

    def fit(self, X, y):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)
        y = self._ensure_output_format(y, sparse_format="csc", enforce_sparse=True)

        self.classifiers_ = []
        self._generate_partition(X, y)
        self._label_count = y.shape[1]

        for i in range(self.model_count_):
            classifier = copy.deepcopy(self.knn_classifier)
            y_subset = self._generate_data_subset(y, self.partition_[i], axis=1)
            if issparse(y_subset) and y_subset.ndim > 1 and y_subset.shape[1] == 1:
                y_subset = np.ravel(y_subset.toarray())
            classifier.fit(
                self._ensure_input_format(X), self._ensure_output_format(y_subset)
            )
            self.classifiers_.append(classifier)

        self.knn_layer_ = copy.deepcopy(self.classifiers_)
        self.classifiers_ = []

        class_membership = self._get_class_membership(self.knn_layer_, X)
        X_concat_class_membership = self._concatenate_class_membership(
            X, class_membership
        )

        for i in range(self.model_count_):
            classifier = copy.deepcopy(self.classifier)
            y_subset = self._generate_data_subset(y, self.partition_[i], axis=1)
            if issparse(y_subset) and y_subset.ndim > 1 and y_subset.shape[1] == 1:
                y_subset = np.ravel(y_subset.toarray())
            classifier.fit(
                self._ensure_input_format(X_concat_class_membership),
                self._ensure_output_format(y_subset),
            )
            self.classifiers_.append(classifier)

        self.classes_ = [classifier.classes_ for classifier in self.classifiers_]

        return self

    def predict(self, X):
        """Predict labels from X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        class_membership = self._get_class_membership(self.knn_layer_, X)
        X_test_concat_membership = self._concatenate_class_membership(
            X, class_membership
        )

        predictions = [
            self._ensure_multi_label_from_single_class(
                self.classifiers_[label].predict(
                    self._ensure_input_format(X_test_concat_membership)
                )
            )
            for label in range(self.model_count_)
        ]

        return hstack(predictions)

    def predict_proba(self, X):
        """Predict probabilities of each labels from given X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """
        class_membership = self._get_class_membership(self.knn_layer_, X)
        X_test_concat_class_membership = self._concatenate_class_membership(
            X, class_membership
        )

        result = lil_matrix((X.shape[0], self._label_count), dtype="float")

        for label_assignment, classifier in zip(self.partition_, self.classifiers_):
            if isinstance(self.classifier, MLClassifierBase):
                # the multilabel classifier should provide a (n_samples, n_labels) matrix
                # we just need to reorder it column wise
                result[:, label_assignment] = classifier.predict_proba(
                    X_test_concat_class_membership
                )
            else:
                # a base classifier for binary relevance returns
                # n_samples x n_classes, where n_classes = [0, 1] - 1 is the probability of
                # the label being assigned
                result[
                    :, label_assignment
                ] = self._ensure_multi_label_from_single_class(
                    classifier.predict_proba(
                        self._ensure_input_format(X_test_concat_class_membership)
                    )
                )[
                    :, 1
                ]  # probability that label is assignedx

        return result



from .functions import label_correlation, estimate_mising_labels, weight_adjacent_matrix, diagonal_matrix_H, diagonal_matrix_Hc, diagonal_matrix_lambda, graph_laplacian_matrix, predictive_matrix, label_bias
import numpy as np


class SMiLE:
    """SMiLE algorithm for multi label with missing labels
    (Semi-supervised multi-label classification using incomplete label information)

    Parameters
    ----------
    s : float, optional, default : 0.5
        Smoothness parameter for class imbalance
    alpha : float, optional, default : 0.35
        Smoothness assumption parameter, ensures similar instances
        having similar predicted output. This parameter balances the
        importance of the two terms of the equation to optimize
    k : int, optional, default : 5
        Neighbours parameter for clustering during the algorithm.
        It will indicate the number of clusters we want to create
        for the k nearest neighbor (kNN)

    Attributes
    ----------
    L : array, [n_labels, n_labels]
        Correlation matrix between labels
    W : array, [n_samples, n_samples]
        Weighted matrix created by kNN for instances
    estimate_matrix : array-like (n_samples, n_labels)
        Label estimation matrix
        y~ic = yiT * L(.,c) if yic == 0
        y~ic = 1 otherwise
    H : array-like (n_samples, n_samples)
        Diagonal matrix indicating if an element of X is labeled or not    
    diagonal_lambda : array-like (n_samples, n_samples)
        Diagonal matrix having the sum of weights of the weighted matrix
    M : array-like (n_samples, n_samples)
        Graph laplacian matrix
    Hc : array-like (n_samples, n_samples)
        Hc = H - (H*1*1t*Ht)/(N)
    P : array-like (n_features, n_labels)
        P = (X*Hc*Xt + alpha*X*M*Xt)-1 * X*Hc*YPred
        R = dxc
    b : array-like (n_labels)
        Label bias as the second item of the equation
        b = ((estimate_matrix - Pt*X)*H*1)/N

    References
    ----------
    If used, please cite the scikit-multilearn library and the relevant paper:

    .. code-block:: bibtex

        @article{TAN2017192,
          title = {Semi-supervised multi-label classification using incomplete label information},
          author = {Qiaoyu Tan and Yanming Yu and Guoxian Yu and Jun Wang},
          journal = {Neurocomputing},
          volume = {260},
          pages = {192-202},
          year = {2017},
          issn = {0925-2312},
          doi = {https://doi.org/10.1016/j.neucom.2017.04.033},
          url = {https://www.sciencedirect.com/science/article/pii/S092523121730704X},
        }
    
    Examples
    --------
    An example use case for SMiLE algorithm:

    .. code-block:: python

        from skmultilearn.missing import SMiLE

        # initialize SMiLE algorithm with parameters
        classifier = SMiLE(s=0.6, alpha=0.4, k=8)

        # train
        classifier.fit(X,y)

        # predict
        prediction = classifier.predict(X)

    """

    def __init__(self, s=0.5, alpha=0.35, k=5):
        """Initialize properties.

        :param s:
        :param alpha:
        :param k:
        :param W:
        :param L:
        :param estimate_matrix:
        :param H:
        :param diagonal_lambda:
        :param M:
        :param Hc:
        :param P:
        :param b:
        """
        self.s = s
        self.alpha = alpha
        self.k = k

        self.W = None
        self.L = None
        self.estimate_matrix = None
        self.H = None
        self.diagonal_lambda = None
        self.M = None
        self.Hc = None
        self.P = None
        self.b = None


    def fit(self, X, y):
        """Fits the model to training data

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Training instances.
        y : array-like, shape=(n_samples, n_labels)
            Training labels.
        """
        X = np.transpose(X)
        y = np.transpose(y)

        self.L = label_correlation(y, self.s)
        self.estimate_matrix = estimate_mising_labels(y, self.L)
        self.H = diagonal_matrix_H(X, y)
        self.Hc = diagonal_matrix_Hc(self.H)
        self.W = weight_adjacent_matrix(X, self.k)
        self.diagonal_lambda = diagonal_matrix_lambda(self.W)
        self.M = graph_laplacian_matrix(self.diagonal_lambda, self.W)
        self.P = predictive_matrix(X, self.Hc, self.M, self.estimate_matrix, self.alpha)
        self.b = label_bias(self.estimate_matrix, self.P, X, self.H)

        return self

    def predict(self, X):
        """Predicts using the model

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Test instances.
        
        Returns
        -------
        predictions : array-like, shape=(n_labels, n_samples)
            Label predictions for the test instances. (As if it was a regression problem range[0,1])
        predictionsNormalized : array-like, shape=(n_labels, n_samples)
            Label predictions
        """
        X = np.transpose(X)
        predictions = np.zeros(shape=[self.b.shape[0], X.shape[1]])
        for i in range(0, X.shape[1]):
            numerator1 = np.zeros(shape=[self.b.shape[0],1])
            numerator = np.array(np.matmul(np.transpose(self.P), X[:, i]))
            for k in range(numerator1.shape[0]):
                numerator1[k,0] = numerator[k]
            prediction = np.add(numerator1, self.b)
            for k in range(prediction.shape[0]):
                predictions[k,i] = prediction[k]
        for i in range(predictions.shape[0]):
            for j in range(predictions.shape[1]):
                if predictions[i,j] > 0.5:
                    predictions[i,j] = 1
                else:
                    predictions[i,j] = 0

        predictions = np.transpose(predictions)
        return predictions
    
    def getParams(self):
        """Returns the parameters of this model
        
        Returns
        -------
        s : float, optional, default : 0.5
            Smoothness parameter for class imbalance
        alpha : float, optional, default : 0.35
            Smoothness assumption parameter, ensures similar instances
            having similar predicted output. This parameter balances the
            importance of the two terms of the equation to optimize
        k : int, optional, default : 5
            Neighbours parameter for clustering during the algorithm.
            It will indicate the number of clusters we want to create
            for the k nearest neighbor (kNN)
        """
        return self.s, self.alpha, self.k

    def setParams(self, s, alpha, k):
        """Sets the parameters of this model
        
        Parameters
        ----------
        s : float, optional, default : 0.5
            Smoothness parameter for class imbalance
        alpha : float, optional, default : 0.35
            Smoothness assumption parameter, ensures similar instances
            having similar predicted output. This parameter balances the
            importance of the two terms of the equation to optimize
        k : int, optional, default : 5
            Neighbours parameter for clustering during the algorithm.
            It will indicate the number of clusters we want to create
            for the k nearest neighbor (kNN)
        """
        self.s = s
        self.alpha = alpha
        self.k = k
        return None



import numpy as np
from sklearn.cluster import KMeans
from numpy.linalg import inv


def label_correlation(y, s):
    """Correlation between labels in a label matrix

    Parameters
    ----------
    y : array-like (n_labels, n_samples)
        Label matrix
    s : float
        Smoothness parameter

    Returns
    -------
    L : array-like (n_labels, n_labels)
        Label correlation matrix
    """
    L = np.zeros(shape=[y.shape[0], y.shape[0]])

    for i in range(0, y.shape[0]):
        yi = sum(y[i,:])
        for j in range(0, y.shape[0]):
            coincidence = 0
            for k in range(0, y.shape[1]):
                if (int(y[i,k]) == int(1)) and (int(y[j,k]) == int(1)):
                    coincidence += 1
            L[i,j] = (coincidence + s)/(yi + 2*s)
    
    return L


def estimate_mising_labels(y, L):
    """Estimation of the missing labels, using the correlation matrix

    Parameters
    ----------
    y : array-like (n_labels, n_samples)
        Label matrix
    L : array-like (n_labels, n_labels)
        Label correlation matrix

    Returns
    -------
    estimate_matrix : array-like (n_labels, n_samples)
        Label estimation matrix
        y~ic = yiT * L(.,c) if yic == 0
        y~ic = 1 otherwise
    """
    estimate_matrix = np.zeros(shape=[y.shape[0],y.shape[1]], dtype=float)
    for j in range(y.shape[0]):
        for i in range(y.shape[1]):
            if int(y[j,i]) == int(0):
                aux = np.dot(np.transpose(y[:, i]), L[:,j])
                estimate_matrix[j,i] = aux
            else:
                estimate_matrix[j,i] = 1
    # Once we have the matrix, normalize the data
    estimate_matrix_copy = np.copy(estimate_matrix)
    for j in range(y.shape[0]):
        for i in range(y.shape[1]):
            if int(y[j,i]) == int(0):
                if np.sum(estimate_matrix_copy[:,i]) != 0:
                    aux = estimate_matrix_copy[j,i]/(np.sum(estimate_matrix_copy[:,i]))
                    estimate_matrix[j,i] = aux
            
    return estimate_matrix


def weight_adjacent_matrix(X, k):
    """Using the kNN algorithm we will use the clusters to get a weight matrix

    Parameters
    ----------
    X : array-like or sparse matrix (n_features, n_samples)
        Data to classify or in this case to make clusters
    k : int
        Number of clusters we want to make
    
    Returns
    -------
    W : array-like (n_samples, n_samples)
        Weighted matrix created from the predictions of kNN
        wij = 1 if xi is in the same cluster as xj
        wij = 0 other case
    """
    kNN = KMeans(n_clusters=k)
    kNN.fit(np.transpose(X))
    predictions = kNN.predict(np.transpose(X))
    W = np.zeros(shape=[X.shape[1], X.shape[1]], dtype=int)
    for i in range(0, X.shape[1]):
        for j in range(0, X.shape[1]):
            if int(predictions[i]) == int(predictions[j]):
                W[i,j] = 1
            else:
                W[i,j] = 0

    return W


def diagonal_matrix_H(X, y):
    """Diagonal matrix that indicates if X is labeled

    Parameters
    ----------
    X : array-like or sparse matrix (n_features, n_samples)
        Data to classify
    y : array-like (n_labels, n_samples)
        Labels of the data

    Returns
    -------
    H : array-like (n_samples, n_samples)
        Diagonal matrix indicating if an element of X is labeled or not
    """

    H = np.zeros(shape=[X.shape[1], X.shape[1]])

    for i in range(0, X.shape[1]):
        if np.sum(y[:, i]) > 0:
            H[i,i] = 1

    return H


def diagonal_matrix_lambda(W):
    """

    Parameters
    ----------
    W : array-like (n_samples, n_samples)
        Weighted matrix

    Returns
    -------
    diagonal_lambda : array-like (n_samples, n_samples)
        Diagonal matrix having the sum of weights of the weighted matrix
    """
    diagonal_lambda = np.zeros(shape=[W.shape[0], W.shape[1]])
    for i in range(0, W.shape[0]):
        diagonal_lambda[i,i] = np.sum(W[i,:])
    
    return diagonal_lambda


def graph_laplacian_matrix(lambda_matrix, W):
    """

    Parameters
    ----------
    lambda_matrix : array-like (n_samples, n_samples)
        Diagonal matrix having the sum of weights of the weighted matrix
    W : array-like (n_samples, n_samples)
        Weighted matrix

    Returns
    -------
    M : array-like (n_samples, n_samples)
        Graph laplacian matrix
    """
    M = np.zeros(shape=[W.shape[0], W.shape[1]])
    M = np.subtract(lambda_matrix, W)
    return M


def diagonal_matrix_Hc(H):
    """

    Parameters
    ----------
    H : array-like (n_samples, n_samples)
        Diagonal matrix indicating if an element of X is labeled or not

    Returns
    -------
    Hc : array-like (n_samples, n_samples)
        Hc = H - (H*1*1t*Ht)/(N)
    """
    Hc = np.zeros(shape = [H.shape[0], H.shape[0]])
    oneVector = np.ones(shape=[H.shape[0],1])
    numerator1 = np.matmul(H, oneVector)
    numerator2 = np.matmul(np.transpose(oneVector), np.transpose(H))
    numerator = np.matmul(numerator1, numerator2)
    product = numerator/H.shape[0]
    Hc = np.subtract(H, product)
    return Hc


def predictive_matrix(X, Hc, M, estimate_matrix, alpha):
    """Predictive matrix that works as the first item of the equation

    Parameters
    ----------
    X : array-like or sparse matrix (n_features, n_samples)
        Data to be classified or trained
    Hc : array-like (n_samples, n_samples)
        Diagonal matrix obtained from H
    M : array-like(n_samples, n_samples)
        Graph laplacian matrix

    Returns
    -------
    P : array-like (n_features, n_labels)
        P = (X*Hc*Xt + alpha*X*M*Xt)-1 * X*Hc*YPred
        R = dxc
    """
    P = np.zeros(shape=[X.shape[0], estimate_matrix.shape[0]])
    numerator1 = np.matmul(X, Hc)
    numerator1 = np.matmul(numerator1, np.transpose(X))
    numerator2 = np.matmul(M, np.transpose(X))
    numerator2 = alpha * np.matmul(X, numerator2)
    numerator = np.add(numerator1, numerator2)
    numerator = inv(numerator)
    numerator2 = np.matmul(X, Hc)
    numerator2 = np.matmul(numerator2, np.transpose(estimate_matrix))
    P = np.matmul(numerator, numerator2)

    return P


def label_bias(estimate_matrix, P, X, H):
    """Label bias that works as the second item of the equation

    Parameters
    ----------
    estimate_matrix : array-like (n_samples, n_samples)
        Diagonal matrix indicating if an element of X is labeled or not
    P : array-like (n_features, n_labels)
        Predictive item
    X : array-like (n_features, n_samples)
        Data to train or test
    H : array-like (n_samples, n_samples)
        Diagonal matrix indicating if an element of X is labeled or not

    Returns
    -------
    b : array-like (n_labels)
        Label bias as the second item of the equation
        b = ((estimate_matrix - Pt*X)*H*1)/N
    """
    b = np.zeros(estimate_matrix.shape[1])
    aux = np.matmul(np.transpose(P), X)
    numerator1 = np.subtract(estimate_matrix, aux)
    oneVector = np.ones(shape=[H.shape[0], 1])
    numerator2 = np.matmul(H, oneVector)
    numerator = np.matmul(numerator1, numerator2)
    b = numerator / H.shape[0]
    return b



import copy
import numpy as np

from scipy.sparse import hstack, issparse, lil_matrix

from ..base.base import MLClassifierBase
from ..base.problem_transformation import ProblemTransformationBase


class ClassificationHeterogeneousFeature(ProblemTransformationBase):
    """Classification with heterogeneous features

    This model is to augment the feature set with extra features which are
    from one for each label in the dataset. The cyclic dependency between
    features and labels is resolved iteratively.

    There are two BR layers, composed of given base classifiers for each layer.
    The first layer is proposed to reproduce heterogeneous features, which will
    be augment to the original feature set. Heterogeneous features are predict
    probabilities produced by base classifier per each label. The second layer
    is to predict each labels with features from first layer.

    Parameters
    ----------
    classifier : :class:`~sklearn.base.BaseEstimator`
        scikit-learn compatible base classifier
    require_dense : [bool, bool], optional
        whether the base classifier requires dense representations
        for input features and classes/labels matrices in fit/predict.
        If value not provided, sparse representations are used if base classifier is
        an instance of :class:`~skmultilearn.base.MLClassifierBase` and dense otherwise.

    Attributes
    ----------
    model_count_ : int
        number of trained models, in this classifier equal to `n_labels`
    partition_ : List[List[int]], shape=(`model_count_`,)
        list of lists of label indexes, used to index the output space matrix, set in :meth:`_generate_partition`
        via :meth:`fit`
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
        list of classifiers trained per partition, set in :meth:`fit`
    first_layer_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
        list of classifiers trained per partition for obtaining heterogeneous feature, set in :meth:`fit`

    References
    ----------
    If used, please cite the scikit-multilearn library and the relevant paper:

    .. code-block:: bibtex

        @inproceedings{godbole2004discriminative,
            title={Discriminative methods for multi-labeled classification},
            author={Godbole, Shantanu and Sarawagi, Sunita},
            booktitle={Pacific-Asia conference on knowledge discovery and data mining},
            pages={22--30},
            year={2004},
            organization={Springer}
        }

    Examples
    --------
    An example use case for Classification with heterogeneous features
    with an :class:`sklearn.svm.SVC` base classifier which supports sparse input:

    .. code-block:: python

        from skmultilearn.problem_transform import ClassificationHeterogeneousFeature
        from sklearn.model_selection import GridSearchCV
        from sklearn.naive_bayes import MultinomialNB
        from sklearn.svm import SVC

        parameters = [
            {
                'classifier': [MultinomialNB()],
                'classifier__alpha': [0.7, 1.0],
            },
            {
                'classifier': [SVC(probability=True)],
                'classifier__kernel': ['rbf', 'linear'],
            },
        ]

        clf = GridSearchCV(ClassificationHeterogeneousFeature(), parameters, scoring='accuracy')
        clf.fit(X, y)

        # Output the best parameters and the corresponding score
        print(clf.best_params_, clf.best_score_)

        # Example output
        # {'classifier': MultinomialNB(alpha=0.7), 'classifier__alpha': 0.7} 0.21
    """

    def __init__(self, classifier=None, require_dense=None):
        super(ClassificationHeterogeneousFeature, self).__init__(
            classifier, require_dense
        )
        self.first_layer_ = []

    def _generate_partition(self, X, y):
        """Partitions the label space into singletons

        Sets `self.partition_` (list of single item lists) and `self.model_count_` (equal to number of labels).

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            not used, only for API compatibility
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `int`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        self.partition_ = list(range(y.shape[1]))
        self.model_count_ = y.shape[1]

    def _concatenate_clm(self, X, class_membership):
        """Concatenate origin features and heterogeneous features into one vector.

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        class_membership : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_labels)
            label predict probabilities from X
        """
        return hstack([X, class_membership]).tolil()

    def _get_class_membership(self, classifiers, X):
        """Get heterogenous features from X based on trained classifiers

        Parameters
        ----------
        classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `model_count`
            list of pretrained classifiers per partition based on X
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        """
        result = lil_matrix((X.shape[0], self._label_count), dtype="float")

        is_ml_classifier = isinstance(self.classifier, MLClassifierBase)
        for label_assignment, classifier in zip(self.partition_, classifiers):
            if is_ml_classifier:
                # the multilabel classifier should provide a (n_samples, n_labels) matrix
                # we just need to reorder it column wise
                result[:, label_assignment] = classifier.predict_proba(X)
            else:
                # a base classifier for binary relevance returns
                # n_samples x n_classes, where n_classes = [0, 1] - 1 is the probability of
                # the label being assigned
                result[
                    :, label_assignment
                ] = self._ensure_multi_label_from_single_class(
                    classifier.predict_proba(self._ensure_input_format(X))
                )[
                    :, 1
                ]  # probability that label is assigned

        return result

    def fit(self, X, y):
        """Fits classifier to training data

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)
        y = self._ensure_output_format(y, sparse_format="csc", enforce_sparse=True)

        self.classifiers_ = []
        self._generate_partition(X, y)
        self._label_count = y.shape[1]

        for i in range(self.model_count_):
            classifier = copy.deepcopy(self.classifier)
            y_subset = self._generate_data_subset(y, self.partition_[i], axis=1)
            if issparse(y_subset) and y_subset.ndim > 1 and y_subset.shape[1] == 1:
                y_subset = np.ravel(y_subset.toarray())
            classifier.fit(
                self._ensure_input_format(X), self._ensure_output_format(y_subset)
            )
            self.classifiers_.append(classifier)

        self.first_layer_ = copy.deepcopy(self.classifiers_)
        self.classifiers_ = []

        class_membership = self._get_class_membership(self.first_layer_, X)
        X_concat_clm = self._concatenate_clm(X, class_membership)

        for i in range(self.model_count_):
            classifier = copy.deepcopy(self.classifier)
            y_subset = self._generate_data_subset(y, self.partition_[i], axis=1)
            if issparse(y_subset) and y_subset.ndim > 1 and y_subset.shape[1] == 1:
                y_subset = np.ravel(y_subset.toarray())
            classifier.fit(
                self._ensure_input_format(X_concat_clm),
                self._ensure_output_format(y_subset),
            )
            self.classifiers_.append(classifier)

        self.classes_ = [classifier.classes_ for classifier in self.classifiers_]

        return self

    def predict(self, X):
        """Predict labels for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        class_membership = self._get_class_membership(self.first_layer_, X)
        X_test_concat_clm = self._concatenate_clm(X, class_membership)

        predictions = [
            self._ensure_multi_label_from_single_class(
                self.classifiers_[label].predict(
                    self._ensure_input_format(X_test_concat_clm)
                )
            )
            for label in range(self.model_count_)
        ]

        return hstack(predictions)

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """
        class_membership = self._get_class_membership(self.first_layer_, X)
        X_test_concat_clm = self._concatenate_clm(X, class_membership)

        result = lil_matrix((X.shape[0], self._label_count), dtype="float")

        is_ml_classifier = isinstance(self.classifier, MLClassifierBase)
        for label_assignment, classifier in zip(self.partition_, self.classifiers_):
            if is_ml_classifier:
                # the multilabel classifier should provide a (n_samples, n_labels) matrix
                # we just need to reorder it column wise
                result[:, label_assignment] = classifier.predict_proba(
                    X_test_concat_clm
                )
            else:
                # a base classifier for binary relevance returns
                # n_samples x n_classes, where n_classes = [0, 1] - 1 is the probability of
                # the label being assigned
                result[
                    :, label_assignment
                ] = self._ensure_multi_label_from_single_class(
                    classifier.predict_proba(
                        self._ensure_input_format(X_test_concat_clm)
                    )
                )[
                    :, 1
                ]  # probability that label is assigned

        return result



import copy
import numpy as np

from scipy.sparse import issparse, hstack
from sklearn.model_selection import GridSearchCV
from sklearn.neighbors import KNeighborsClassifier

from ..base.problem_transformation import ProblemTransformationBase


class StructuredGridSearchCV(ProblemTransformationBase):
    """Hyperparameter tuning per each label classifier

    As original GridSearchCV provided by scikit-learn ignores
    BR&CC structural property, it cannot search best parameter and classifier
    for each label. Therefore, StructuredGridSearchCV was implemented for fine tuning
    with considering structural property. StructuredGridSearchCV searches best classifier
    with optimal hyper-parameters for each labels.

    StructuredGridSearchCV provides "fit", "predict", "predict_proba" as its methods.
    It provides list of optimal classifiers with fine-tuned hyper-parameters
    via find_optm_classifier function.

    If print_best_param is True, find_optm_classifier function prints
    best parameter for each label.

    Parameters
    ----------
    Same as GridsearchCV in scikit-learn
    print_best_param : bool, default = False
        whether print best parameter each label classifier or not

    Attributes
    ----------
    classifiers_ : List[:class:`~sklearn.base.BaseEstimator`] of shape `n_labels`
        list of classifiers trained per partition, set in :meth:`fit`
    print_best_param : bool, default = False
        whether print best parameter each label classifier or not
    estimator : estimator object.
        This is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.
    classifier_num : int
        The index which corresponds to each label classifier

    Examples
    --------
    An example use case for Structured GridSearchCV:

    .. code-block:: python

        from skmultilearn.problem_transform import StructuredGridSearchCV, BinaryRelevance
        from sklearn.naive_bayes import MultinomialNB

        # Define parameter grid
        parameters = {'classifier': MultinomialNB(), 'alpha':[1, 10]}

        # Create StructuredGridSearchCV instance
        structured_grid_search = StructuredGridSearchCV(
            estimator=BinaryRelevance(),
            param_grid=parameters,
            scoring='accuracy',
            print_best_param=True
        )

        # Fit and find optimal classifiers
        structured_grid_search.fit(X_train, y_train)
        optimal_classifiers = structured_grid_search.find_optm_classifiers(X, y)
    """

    def __init__(
        self,
        estimator,
        param_grid,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=None,
        verbose=0,
        pre_dispatch="2*n_jobs",
        error_score=np.nan,
        return_train_score=False,
        print_best_param=False,
    ):
        super(StructuredGridSearchCV, self).__init__()
        self.classifiers_ = []
        self.print_best_param = print_best_param
        self.estimator = estimator
        self.classifier_num = 0

        if isinstance(param_grid, list):
            raise NotImplementedError(
                "Handling multiple classifiers has not yet been implemented for StructuredGridSearchCV. Feel free to open a PR to fix it."
            )

        # Temporary fix until handling of multiple classifiers in param_grid is added
        self.estimator.classifier = (
            param_grid["classifier"][0]
            if isinstance(param_grid["classifier"], list)
            and len(param_grid["classifier"]) > 0
            else param_grid["classifier"]
        )
        del param_grid["classifier"]

        if type(self.estimator).__name__ == "InstanceBasedLogisticRegression":
            self.param_grid = param_grid
            param_grid = {"n_neighbors": [10]}
            self.gsc = GridSearchCV(
                estimator=KNeighborsClassifier(),
                param_grid=param_grid,
                scoring=scoring,
                n_jobs=n_jobs,
                refit=refit,
                cv=cv,
                verbose=verbose,
                pre_dispatch=pre_dispatch,
                error_score=error_score,
                return_train_score=return_train_score,
            )

        else:
            self.gsc = GridSearchCV(
                estimator=estimator.classifier,
                param_grid=param_grid,
                scoring=scoring,
                n_jobs=n_jobs,
                refit=refit,
                cv=cv,
                verbose=verbose,
                pre_dispatch=pre_dispatch,
                error_score=error_score,
                return_train_score=return_train_score,
            )

    def fit(self, X, y):
        """Fits classifier to training data and finds optimal classifier

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        self
            fitted instance of self

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        self.estimator.fit(X, y)

        if type(self.estimator).__name__ == "ClassificationHeterogeneousFeature":
            self.estimator.first_layer_ = self.find_optm_classifiers(X, y)

            class_membership = self.estimator._get_class_membership(
                self.estimator.first_layer_, X
            )
            X_concat_clm = self.estimator._concatenate_clm(X, class_membership)
            self.classifiers_ = self.find_optm_classifiers(X_concat_clm, y)

        elif type(self.estimator).__name__ == "InstanceBasedLogisticRegression":
            self.estimator.knn_layer = self.find_optm_classifiers(X, y)

            class_membership = self.estimator._get_class_membership(
                self.estimator.knn_layer, X
            )
            X_concat_clm = self.estimator._concatenate_class_membership(
                X, class_membership
            )
            self.gsc = GridSearchCV(
                estimator=self.estimator.classifier, param_grid=self.param_grid
            )
            self.classifiers_ = self.find_optm_classifiers(X_concat_clm, y)

        else:
            self.classifiers_ = self.find_optm_classifiers(X, y)

        return self

    def predict(self, X):
        """Predict labels for X using optimal classifiers

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        self.estimator.classifiers_ = self.classifiers_
        return self.estimator.predict(self._ensure_input_format(X))

    def predict_proba(self, X):
        """Predict probabilities of label assignments for X using optimal classifiers

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix

        Returns
        -------
        :mod:`scipy.sparse` matrix of `float in [0.0, 1.0]`, shape=(n_samples, n_labels)
            matrix with label assignment probabilities
        """
        self.estimator.classifiers_ = self.classifiers_
        return self.estimator.predict_proba(self._ensure_input_format(X))

    def find_optm_classifiers(self, X, y):
        """Find optimal classifier per label

        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments

        Returns
        -------
        optimized_clfs_
            list of optimal classifier per label

        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        optimized_clfs_ = []
        if type(self.estimator).__name__ == "ClassifierChain":
            optimized_clfs_ = [None for x in range(y.shape[1])]
            for i in self.estimator._order():
                gridsearchCV_ = copy.deepcopy(self.gsc)
                y_subset = self._generate_data_subset(y, i, axis=1)
                gridsearchCV_.fit(
                    self._ensure_input_format(X), self._ensure_output_format(y_subset)
                )
                if self.print_best_param:
                    print(
                        "[",
                        (self.classifier_num + 1),
                        "]",
                        " Classifier Best Parameters : ",
                        gridsearchCV_.best_params_,
                    )
                optimized_clfs_[i] = gridsearchCV_.best_estimator_
                X = hstack([X, y_subset])
                self.classifier_num += 1
        else:
            for i in range(len(self.estimator.classifiers_)):
                gridsearchCV_ = copy.deepcopy(self.gsc)
                y_subset = self._generate_data_subset(y, i, axis=1)
                if issparse(y_subset) and y_subset.ndim > 1 and y_subset.shape[1] == 1:
                    y_subset = np.ravel(y_subset.toarray())
                gridsearchCV_.fit(
                    self._ensure_input_format(X), self._ensure_output_format(y_subset)
                )
                if self.print_best_param:
                    print(
                        "[",
                        (self.classifier_num + 1),
                        "]",
                        " Classifier Best Parameters : ",
                        gridsearchCV_.best_params_,
                    )
                optimized_clf = gridsearchCV_.best_estimator_
                optimized_clfs_.append(optimized_clf)
                self.classifier_num += 1
        return optimized_clfs_



import numpy as np
from scipy.sparse import issparse
from sklearn.base import BaseEstimator, ClassifierMixin, clone
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from abc import ABC, abstractmethod


class SplitCriterion(ABC):
    @abstractmethod
    def calculate_impurity(self, labels):
        """Calculate the impurity of a dataset."""
        raise NotImplementedError

    @abstractmethod
    def calculate_gain(self, base_impurity, left_labels, right_labels):
        """Calculate the information gain from a split."""
        raise NotImplementedError


class GiniCriterion(SplitCriterion):
    def calculate_impurity(self, labels):
        if labels.size == 0:
            return 0
        if issparse(labels):
            label_sums = labels.sum(axis=0).A1
        else:
            label_sums = np.sum(labels, axis=0)
        total_label_count = np.sum(label_sums)
        if total_label_count == 0:
            return np.inf
        label_probs = label_sums / total_label_count
        return 1 - np.sum(np.multiply(label_probs, label_probs))

    def calculate_gain(self, base_impurity, left_labels, right_labels):
        left_impurity = self.calculate_impurity(left_labels)
        right_impurity = self.calculate_impurity(right_labels)
        weight_left = left_labels.shape[0] / (
            left_labels.shape[0] + right_labels.shape[0]
        )
        weight_right = 1 - weight_left
        return base_impurity - (
            weight_left * left_impurity + weight_right * right_impurity
        )


class EntropyCriterion(SplitCriterion):
    def calculate_impurity(self, labels):
        if issparse(labels):
            if not labels.nnz:
                return 0
            label_sums = labels.sum(axis=0).A1
        else:
            if not labels.any():
                return 0
            label_sums = np.sum(labels, axis=0)
        total = label_sums.sum()
        if total == 0:
            return 0
        label_probs = label_sums / total
        label_probs = label_probs[label_probs > 0]
        return -np.sum(label_probs * np.log2(label_probs))

    def calculate_gain(self, base_impurity, left_labels, right_labels):
        left_entropy = self.calculate_impurity(left_labels)
        right_entropy = self.calculate_impurity(right_labels)
        total = left_labels.shape[0] + right_labels.shape[0]
        weighted_avg_entropy = (
            left_labels.shape[0] * left_entropy + right_labels.shape[0] * right_entropy
        ) / total
        return base_impurity - weighted_avg_entropy


class CorrelationCriterion(SplitCriterion):
    def calculate_impurity(self, labels):
        if labels.size == 0:
            return np.inf
        if issparse(labels):
            labels = labels.toarray()
        if np.all(np.all(labels == labels[0, :], axis=0)):
            return np.inf
        std_labels = np.std(labels, axis=0)
        valid_cols = std_labels > 0
        if np.sum(valid_cols) < 2:
            return np.inf
        valid_labels = labels[:, valid_cols]
        corr_matrix = np.abs(np.corrcoef(valid_labels, rowvar=False))
        np.fill_diagonal(corr_matrix, 0)
        avg_corr = np.mean(corr_matrix)
        return avg_corr

    def calculate_gain(self, base_impurity, left_labels, right_labels):
        left_impurity = self.calculate_impurity(left_labels)
        right_impurity = self.calculate_impurity(right_labels)
        total_samples = left_labels.shape[0] + right_labels.shape[0]
        weighted_avg_impurity = (
            left_labels.shape[0] * left_impurity
            + right_labels.shape[0] * right_impurity
        ) / total_samples
        return base_impurity - weighted_avg_impurity


class PredictiveClusteringTree(BaseEstimator, ClassifierMixin):
    """
    A predictive clustering tree (PCT) algorithm for multi-label classification that supports multiple split criteria.

    This algorithm constructs a decision tree where each leaf node represents a multi-label classifier trained on a
    subset of the data. It partitions the feature space recursively, aiming to find splits that lead to optimal
    separation based on a specified impurity criterion, thereby potentially capturing complex label dependencies more
    effectively.

    The flexibility to choose between different splitting criteria (e.g., Gini, entropy, correlation) allows for tailored
    approaches to handling multi-label data, enabling the algorithm to better accommodate the specific characteristics
    and correlations present in the labels.

    Parameters
    ----------
    classifier : estimator, default=DecisionTreeClassifier()
        The base classifier used at each leaf node of the tree. This classifier is trained on the subsets of data
        determined by the tree splits.¨
    criterion : SplitCriterion instance, default=GiniCriterion()
        The criterion used to evaluate splits. Must be an instance of a class that extends the SplitCriterion
        abstract base class.
    max_depth : int, default=5
        The maximum depth of the tree. Limits the number of recursive splits to prevent overfitting.
    min_samples_split : int, default=2
        The minimum number of samples required to consider splitting an internal node. Helps prevent creating nodes
        with too few samples.
    min_samples_leaf : int, default=1
        The minimum number of samples a leaf node must have. Ensures that each leaf has a minimum size,impacting the
        granularity of the model.

    Attributes
    ----------
    n_features_in_ : int
        The number of features in the input data upon fitting the model.
    tree_ : Node
        The root node of the decision tree. Each node in the tree represents a decision point or a leaf with an
        associated classifier.

    Methods
    -------
    fit(X, y):
        Fit the predictive clustering tree model to the training data.
    predict(X):
        Predict multi-label outputs for the input data using the trained tree.

    Notes
    -----
    .. note ::

        The tree-building process relies heavily on the chosen split criterion's ability to evaluate and select the most
        informative splits. Custom split criteria can be implemented by extending the SplitCriterion abstract base class.

    .. note ::

        Currenlty only dense input data is supported.

    Examples
    --------
    .. code-block:: python

        from sklearn.datasets import make_multilabel_classification
        X, y = make_multilabel_classification(n_samples=100, n_features=20, n_classes=3, n_labels=2, random_state=42)
        pct = PredictiveClusteringTree(criterion=GiniCriterion(), max_depth=4, min_samples_split=2, min_samples_leaf=1)
        pct.fit(X, y)
        pct.predict(X[0:5])
    """

    def __init__(
        self,
        classifier=DecisionTreeClassifier(),
        criterion=GiniCriterion(),
        max_depth=5,
        min_samples_split=2,
        min_samples_leaf=1,
    ):
        self.criterion = criterion
        self.classifier = classifier
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf

    class Node:
        """
        Inner class representing a node in the predictive clustering tree.

        Attributes
        ----------
        feature_index : int or None
            The index of the feature used for splitting at this node.
        threshold : float or None
            The threshold value for the split.
        left : Node or None
            The left child node.
        right : Node or None
            The right child node.
        classifier : estimator or None
            The classifier associated with the leaf node.
        """

        def __init__(self):
            self.feature_index = None
            self.threshold = None
            self.left = None
            self.right = None
            self.classifier = None

    def fit(self, X, y):
        """
        Fit the predictive clustering tree to the training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input feature matrix.
        y : array-like of shape (n_samples, n_labels)
            The binary indicator matrix with label assignments.

        Returns
        -------
        self : object
            The fitted instance of the classifier.
        """
        X, y = check_X_y(X, y, multi_output=True, accept_sparse=True)
        self.n_features_in_ = X.shape[1]
        self.tree_ = self._grow_tree(X, y)
        return self

    def _check_all_rows_same(self, y):
        if issparse(y):
            if y.shape[0] <= 1:
                return True
            first_row = y[0].toarray()
            for i in range(1, y.shape[0]):
                if not np.array_equal(first_row, y[i].toarray()):
                    return False
            return True
        else:
            return len(np.unique(y, axis=0)) == 1

    def _grow_tree(self, X, y, depth=0):
        """
        Recursively grows the predictive clustering tree.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input feature matrix.
        y : array-like of shape (n_samples, n_labels)
            The binary indicator matrix with label assignments.
        depth : int, optional
            The current depth of the tree (default is 0).

        Returns
        -------
        node : Node
            The root node of the subtree.
        """
        if issparse(X):
            X = X.tocsc()
        base_impurity = self.criterion.calculate_impurity(y)
        all_rows_not_same = self._check_all_rows_same(y)
        if (
            all_rows_not_same
            or depth >= self.max_depth
            or X.shape[0] < self.min_samples_split
        ):
            node = self.Node()
            dense_y = y.toarray() if issparse(y) else y
            node.classifier = clone(self.classifier).fit(X, dense_y)
            return node

        best_gain = -np.inf
        best_idx, best_thr = None, None

        for idx in range(self.n_features_in_):
            if issparse(X):
                column_values = X[:, idx].toarray().ravel()
            else:
                column_values = X[:, idx]
            thresholds = np.unique(column_values)
            for thr in thresholds:
                if issparse(X):
                    left_idx = X[:, idx].toarray().ravel() < thr
                    right_idx = np.logical_not(left_idx)
                else:
                    left_idx = column_values < thr
                    right_idx = ~left_idx

                if np.sum(left_idx) >= self.min_samples_leaf and np.sum(right_idx) >= self.min_samples_leaf:
                    y_left, y_right = y[left_idx], y[right_idx]
                    gain = self.criterion.calculate_gain(base_impurity, y_left, y_right)
                    if gain != np.inf and gain > best_gain:
                        best_gain = gain
                        best_idx, best_thr = idx, thr

        if best_idx is not None:
            if issparse(X):
                left_idx = X[:, best_idx].toarray().ravel() < best_thr
                right_idx = np.logical_not(left_idx)
            else:
                left_idx = X[:, best_idx] < best_thr
                right_idx = ~left_idx

            X_left, y_left = X[left_idx], y[left_idx]
            X_right, y_right = X[right_idx], y[right_idx]
            node = self.Node()
            node.feature_index = best_idx
            node.threshold = best_thr
            node.left = self._grow_tree(X_left, y_left, depth + 1)
            node.right = self._grow_tree(X_right, y_right, depth + 1)
            return node
        else:
            node = self.Node()
            dense_y = y.toarray() if issparse(y) else y
            node.classifier = clone(self.classifier).fit(X, dense_y)
            return node

    def predict(self, X):
        """
        Predict multi-label outputs for the input data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input feature matrix.

        Returns
        -------
        predictions : array-like of shape (n_samples, n_labels)
            The binary indicator matrix with predicted label assignments.
        """
        check_is_fitted(self, ["tree_", "n_features_in_"])
        X = check_array(X, accept_sparse=True)
        if issparse(X):
            predictions = [self._predict(X[i].toarray().ravel(), self.tree_) for i in range(X.shape[0])]
        else:
            predictions = [self._predict(inputs, self.tree_) for inputs in X]

        return np.array(predictions)

    def _predict(self, inputs, node):
        """
        Recursively predicts labels for the input data.

        Parameters
        ----------
        inputs : array-like of shape (n_features,)
            The input feature vector.
        node : Node
            The current node in the tree.

        Returns
        -------
        prediction : array-like of shape (n_labels,)
            The predicted label assignments.
        """
        if node.classifier:
            return node.classifier.predict([inputs])[0]
        elif inputs[node.feature_index] < node.threshold:
            return self._predict(inputs, node.left)
        else:
            return self._predict(inputs, node.right)



"""
The :mod:`skmultilearn.problem_transform` module provides classifiers
that follow the problem transformation approaches to multi-label classification.

The problem transformation approach to multi-label classification converts multi-label problems to
single-label problems: single-class or multi-class.


+-----------------------------------------------------------------------------+------------------------------------------------+
| Classifier                                                                  | Description                                    |
+=============================================================================+================================================+
| :class:`~skmultilearn.problem_transform.BinaryRelevance`                    |  treats each label as a separate single-class  |
|                                                                             |  classification problem                        |
+-----------------------------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.ClassifierChain`                    |  treats each label as a part of a conditioned  |
|                                                                             |  chain of single-class classification problems |
+-----------------------------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.ProbabilisticClassifierChain`       |  extends Classifier Chains by modeling joint   |
|                                                                             |  label distributions and estimating the        |
|                                                                             |  probability of label sets. It trains a series |
|                                                                             |  of classifiers, each predicting the           |
|                                                                             |  probability of a label, conditioned on the    |
|                                                                             |  input features and preceding label predictions|
+-----------------------------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.ClassificationHeterogeneousFeature` | augments the feature set                       |
|                                                                             | with extra features derived from label         |
|                                                                             | probabilities and resolves cyclic dependencies |
|                                                                             | between features and labels iteratively        |
+-----------------------------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.LabelPowerset`                      | treats each label combination as a separate    |
|                                                                             | class with one multi-class classification      |
|                                                                             | problem                                        |
+-----------------------------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.InstanceBasedLogisticRegression`    | combines instance-based learning with logistic |
|                                                                             | regression, using neighbors' information as    |
|                                                                             | features. It has a K-Nearest Neighbor layer    |
|                                                                             | followed by Logistic Regression classifiers.   |
+-----------------------------------------------------------------------------+------------------------------------------------+
| :class:`~skmultilearn.problem_transform.StructuredGridSearchCV`             | performs hyperparameter tuning for each label  |
|                                                                             | classifier, considering BR&CC structural       |
|                                                                             | properties. It searches for optimal classifiers|
|                                                                             | with fine-tuned parameters for each label.     |
+-----------------------------------------------------------------------------+------------------------------------------------+

"""

from .br import BinaryRelevance
from .cc import ClassifierChain
from .pcc import ProbabilisticClassifierChain
from .chf import ClassificationHeterogeneousFeature
from .gsc import StructuredGridSearchCV
from .lp import LabelPowerset
from .iblr import InstanceBasedLogisticRegression


__all__ = [
    "BinaryRelevance",
    "ClassifierChain",
    "ProbabilisticClassifierChain",
    "ClassificationHeterogeneousFeature",
    "LabelPowerset",
    "InstanceBasedLogisticRegression",
    "StructuredGridSearchCV",
]
